; Author:      Roy Klein Gebbinck (www.b-k-g.nl)
; Version:     20180329: First version.
; Created for: https://forum.bricsys.com/discussion/33560/how-can-i-pick-up-a-single-vertex-from-a-3d-solid-and-move-it

; Return value: Non-empty pickset or nil.
; Erased objects are not added to the set.
(defun KGA_Conv_ObjectList_To_Pickset (lst / ret)
  (setq ret (ssadd))
  (foreach obj lst (if (not (vlax-erased-p obj)) (ssadd (vlax-vla-object->ename obj) ret)))
  (if (/= 0 (sslength ret)) ret)
)

(defun KGA_Sys_ActiveCommand_P ()
  (/= (logand (getvar 'cmdactive) 3) 0)
)

(defun KGA_Sys_ObjectOwner (obj)
  (vla-objectidtoobject (vla-get-database obj) (vla-get-ownerid obj))
)

; 20180329
(defun BKG_PfmToSol_CreateFaces (obj / lyr newLst nme)
  (while (tblsearch "layer" (setq nme (strcat "BKG_PfmToSol_" (rtos (getvar 'cdate) 2 8))))
  ) ; Temp name for layer.
  (setq lyr (vla-add (vla-get-layers (vla-get-activedocument (vlax-get-acad-object))) nme))
  (vla-put-layer obj nme)
  (setvar 'cmdecho 0)
  (command "_.explode" (vlax-vla-object->ename obj))
  (setvar 'cmdecho 1)
  (setq newLst
    (mapcar
      '(lambda (enm / obj)
        (setq obj (vlax-ename->vla-object enm))
        (vla-put-layer obj (getvar 'clayer))
        obj
      )
      (vle-selectionset->list (ssget "_X" (list (cons 8 nme))))
    )
  )
  (vla-delete lyr)
  newLst
)

; 20180329
(defun BKG_PfmToSol_CreateRegions (objLst / spc)
  (setq spc (KGA_Sys_ObjectOwner (car objLst)))
  (mapcar
    '(lambda (obj / new)
      (setq new (car (vlax-invoke spc 'addregion (list obj))))
      (vla-delete obj)
      new
    )
    objLst
  )
)

; 20180329
(defun BKG_PfmToSol_CreateSolid (objLst / las ss)
  (if
    (and
      (setq ss (KGA_Conv_ObjectList_To_Pickset objLst))
      (setq las (entlast))
      (setvar 'cmdecho 0)
      (not (command "_.dmstitch" ss ""))
      (not (equal las (setq las (entlast))))
      (not (command "_.dmsimplify" las ""))
      (setvar 'cmdecho 1)
    )
    (vlax-ename->vla-object las)
  )
)

; 20180329
(defun c:BKG_PfmToSol ( / doc enm faceLst new obj regionLst sol)
  (setq doc (vla-get-activedocument (vlax-get-acad-object)))
  (vla-endundomark doc)
  (vla-startundomark doc)
  (if
    (and
      (or
        (and
          (getcname "dmstitch")
          (getcname "dmsimplify")
        )
        (prompt "\nError: required commands _DmStitch and/or _DmSimplify are not available ")
      )
      (or
        (not (KGA_Sys_ActiveCommand_P))
        (prompt "\nError: command cannot be used transparently ")
      )
      (or
        (= (logand (cdr (assoc 70 (tblsearch "layer" (getvar 'clayer)))) 4) 0)
        (prompt "\nError: current layer is locked ")
      )
      (setq enm (car (entsel "\nSelect polyface mesh: ")))
      (setq obj (vlax-ename->vla-object enm))
      (or
        (= "AcDbPolyFaceMesh" (vla-get-objectname obj))
        (prompt "\nError: object is not a polyface mesh ")
      )
      (or
        (= (logand (cdr (assoc 70 (tblsearch "layer" (vla-get-layer obj)))) 4) 0)
        (prompt "\nError: object layer is locked ")
      )
      (setq new (vla-copy obj))
      (setq faceLst (BKG_PfmToSol_CreateFaces new))
      (setq regionLst (BKG_PfmToSol_CreateRegions faceLst))
      (setq sol (BKG_PfmToSol_CreateSolid regionLst))
    )
    (progn
      (foreach prop '(color layer linetype linetypescale lineweight material plotstylename truecolor)
        (vlax-put sol prop (vlax-get obj prop))
      )
      (vla-delete obj)
      (sssetfirst nil (ssadd (vlax-vla-object->ename sol)))
    )
  )
  (vla-endundomark doc)
  (princ)
)
